<?php

declare(strict_types=1);

// cspell:ignore imagebmp
namespace Drupal\Tests\migrate_devel_file_copy\Kernel\Plugin\migrate\process;

use Drupal\Tests\migrate\Kernel\process\FileCopyTest;
use Drupal\migrate\MigrateException;
use Drupal\migrate\MigrateExecutableInterface;
use Drupal\migrate\Plugin\MigrateProcessInterface;
use Drupal\migrate\Row;
use Drupal\migrate_devel_file_copy\Plugin\migrate\process\FileCopyOrGenerate;

/**
 * Tests the file_copy_or_generate process plugin.
 *
 * @group migrate_devel_file_copy
 *
 * @coversDefaultClass \Drupal\migrate_devel_file_copy\Plugin\migrate\process\FileCopyOrGenerate
 */
class FileCopyOrGenerateTest extends FileCopyTest {

  /**
   * {@inheritdoc}
   */
  public function testDownloadRemoteUri(): void {
    $executable = $this->prophesize(MigrateExecutableInterface::class)->reveal();
    $row = $this->prophesize(Row::class)->reveal();
    $download_plugin = $this->createMock(MigrateProcessInterface::class);
    $download_plugin
      ->expects($this->once())
      ->method('transform');

    $plugin = new FileCopyOrGenerate(
      [],
      'file_copy_or_generate',
      [],
      $this->container->get('stream_wrapper_manager'),
      $this->container->get('file_system'),
      $download_plugin
    );

    $plugin->transform(
      [
        'https://drupal.org/favicon.ico',
        '/destination/path',
      ],
      $executable,
      $row,
      'destination_property'
    );
  }

  /**
   * Tests that missing files are generated.
   *
   * @param string $source
   *   The source of the missing file.
   * @param string $destination
   *   The destination path of the missing file.
   * @param string $expected_mime
   *   The expected mime type of the generated file.
   *
   * @dataProvider providerTestNonExistentSourceFile
   */
  public function testNonExistentSourceFile(
    string $source = '',
    string $destination = '',
    string $expected_mime = '',
  ): void {
    $actual_destination = $this->doTransform($source, $destination);
    $this->assertFileExists($actual_destination);
    $this->assertSame($actual_destination, $destination, "The process plugin wasn't able to generate the file");
    if (extension_loaded('fileinfo')) {
      $actual_mime_type = finfo_file(
        finfo_open(FILEINFO_MIME_TYPE),
        $actual_destination
      );
      $this->assertEquals($expected_mime, $actual_mime_type);
    }
  }

  /**
   * Data provider for ::testNonExistentSourceFile.
   *
   * @return array[]
   *   The test cases.
   */
  public static function providerTestNonExistentSourceFile(): array {
    $bmpMime = 'text/plain';
    if (function_exists('imagebmp')) {
      $bmpMime = version_compare(PHP_VERSION, '8.3', 'ge')
        ? 'image/bmp'
        : 'image/x-ms-bmp';
    }
    return [
      'jpg' => [
        'source' => '/missing/image.jpg',
        'destination' => 'public://existing/image.jpg',
        'expected_mime' => 'image/jpeg',
      ],
      'jpeg' => [
        'source' => '/missing/image.jpeg',
        'destination' => 'public://existing/image.jpeg',
        'expected_mime' => 'image/jpeg',
      ],
      'JPG' => [
        'source' => '/missing/iMaGe.JPG',
        'destination' => 'public://existing/iMaGe.JPG',
        'expected_mime' => 'image/jpeg',
      ],
      'gif' => [
        'source' => '/missing/image.gif',
        'destination' => 'public://existing/image.gif',
        'expected_mime' => 'image/gif',
      ],
      'png' => [
        'source' => '/missing/image.png',
        'destination' => 'public://existing/image.png',
        'expected_mime' => 'image/png',
      ],
      'bmp' => [
        'source' => '/missing/image.bmp',
        'destination' => 'public://existing/image.bmp',
        'expected_mime' => $bmpMime,
      ],
      'webp' => [
        'source' => '/missing/image.webp',
        'destination' => 'public://existing/image.webp',
        'expected_mime' => 'text/plain',
      ],
      'txt' => [
        'source' => '/missing/file.txt',
        'destination' => 'public://existing/file.txt',
        'expected_mime' => 'text/plain',
      ],
      'pdf' => [
        'source' => '/missing/file.pdf',
        'destination' => 'public://existing/file.pdf',
        'expected_mime' => 'text/plain',
      ],
      'zip' => [
        'source' => '/missing/file.zip',
        'destination' => 'public://existing/file.zip',
        'expected_mime' => 'text/plain',
      ],
      'gz' => [
        'source' => '/missing/file.gz',
        'destination' => 'public://existing/file.gz',
        'expected_mime' => 'text/plain',
      ],
      'no extension' => [
        'source' => '/missing/file',
        'destination' => 'public://existing/file',
        'expected_mime' => 'text/plain',
      ],
      'Missing remote txt' => [
        'source' => 'http://localhost/missing/file.txt',
        'destination' => 'public://existing/file.txt',
        'expected_mime' => 'text/plain',
      ],
    ];
  }

  /**
   * {@inheritdoc}
   *
   * No need for verifying the download plugin's reuse: that should be tested by
   * the download plugin's test.
   *
   * @see providerSuccessfulReuse
   */
  public static function providerSuccessfulReuseStatic(): array {
    return [
      'Reusing a local file' => [
        'source_path' => static::getDrupalRoot() . '/core/tests/fixtures/files/image-test.jpg',
        'destination_path' => 'public://file1.jpg',
      ],
    ];
  }

  /**
   * {@inheritdoc}
   *
   * @dataProvider providerSuccessfulReuseStatic
   */
  public function testSuccessfulReuse($source_path, $destination_path): void {
    $file_reuse = $this->doTransform($source_path, $destination_path);
    clearstatcache(TRUE, $destination_path);
    $timestamp = time() - 10;
    touch($file_reuse, $timestamp);

    // We need to make sure the modified timestamp on the file is sooner than
    // the attempted migration.
    $configuration = ['file_exists' => 'use existing'];
    $this->doTransform($source_path, $destination_path, $configuration);
    clearstatcache(TRUE, $destination_path);
    $this->assertEquals(
      $timestamp,
      (new \SplFileInfo($destination_path))->getMTime(),
    );

    $this->doTransform($source_path, $destination_path);
    clearstatcache(TRUE, $destination_path);
    $this->assertGreaterThan(
      $timestamp,
      (new \SplFileInfo($destination_path))->getMTime(),
    );
  }

  /**
   * {@inheritdoc}
   */
  protected function doTransform($source_path, $destination_path, $configuration = [], bool $expect_download = FALSE): string {
    $executable = $this->prophesize(MigrateExecutableInterface::class)->reveal();
    $row = $this->prophesize(Row::class)->reveal();
    $download_plugin = $this->createMock(MigrateProcessInterface::class);
    if ($expect_download) {
      $download_plugin
        ->expects($this->once())
        ->method('transform');
    }
    else {
      $download_plugin
        ->method('transform')
        ->will($this->returnCallback(function ($value) {
          throw new MigrateException(sprintf(
            'Client error: ... (%s)',
            $value[0]
          ));
        }));
    }

    $plugin = new FileCopyOrGenerate(
      $configuration,
      'file_copy_or_generate',
      [],
      $this->container->get('stream_wrapper_manager'),
      $this->container->get('file_system'),
      $download_plugin
    );

    return $plugin->transform(
      [
        $source_path,
        $destination_path,
      ],
      $executable,
      $row,
      'destination_property'
    );
  }

}
